import argparse
import socket
import time
import logging
import json
import traceback
import os

from .engine import QueueIsFull, TaskIdNotFound, TaskNotCompleted, ActivityData
from .messages import *

logger = logging.getLogger('adengine.client')

ADENGINE_UNIX_SOCK = os.environ.get('ADENGINE_UNIX_SOCK') or DEFAULT_UNIX_SOCKET


class Client:
    """Class representing client for ADEngine server."""
    def __init__(self, unix_socket: str):
        self._unix_socket = unix_socket

    def put(self, activity: bytes, password: str or None) -> str:
        """Puts task to queue."""
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self._unix_socket)
        send(sock, TaskMessage(activity, password))
        message = recv(sock)
        sock.close()

        if isinstance(message, QueueIsFullMessage):
            raise QueueIsFull
        elif isinstance(message, TaskIdMessage):
            return message.task_id

    def get(self, task_id: str) -> ActivityData:
        """Gets task result."""
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self._unix_socket)
        send(sock, TaskIdMessage(task_id))
        message = recv(sock)
        sock.close()

        if isinstance(message, TaskIdNotFoundMessage):
            raise TaskIdNotFound
        elif isinstance(message, TaskNotCompletedMessage):
            raise TaskNotCompleted
        elif isinstance(message, ResultMessage):
            return message.activity_data


def parse_args() -> dict:
    parser = argparse.ArgumentParser('ADEngine client')
    parser.add_argument('activity',
                        help='Path to activity file')
    parser.add_argument('-p', '--password',
                        help='Password for activity file',
                        dest='password',
                        default=None)
    parser.add_argument('-ux', '--unix_socket',
                        help='Path to unix socket',
                        dest='unix_socket',
                        default=ADENGINE_UNIX_SOCK)
    return vars(parser.parse_args())


def main() -> int:
    args = parse_args()
    client = Client(args['unix_socket'])

    with open(args['activity'], 'rb') as file:
        activity = file.read()

    try:
        try:
            task_id = client.put(activity, args['password'])
        except QueueIsFull:
            logger.info('Queue is full')
            return 1
        else:
            result = None
            while not result:
                try:
                    result = client.get(task_id)
                except TaskNotCompleted:
                    logger.info('Task not ready')
                    time.sleep(0.5)
                    continue
                else:
                    logger.info(f'Task completed!')

                    if result.error:
                        logger.info(f'Error: {result.error.__class__.__name__}')
                    else:
                        data = {
                            'consumed_time': str(result.consumed_time),
                            'data': result.data,
                        }
                        logger.info(json.dumps(data, indent=4))
    except Exception as e:
        logger.error(f'An error occurred :\n'
                     f'{traceback.format_exception(e.__class__, e, e.__traceback__)}')
        return 1

    return 0


if __name__ == '__main__':
    exit(main())



import pytest
import os
import time
import datetime
import threading
import typing

from adengine.engine import *
from adengine.pt import *

from test import TEST_DATA, USE_VIRTUAL_DISPLAY, read_file_binary

QUEUE_SIZE = 10
WORKERS_NUM = 2
ADE_TIMEOUT = 5
RESULT_TTL = datetime.timedelta(hours=1)


@pytest.fixture(scope='module')
def launch_engine():
    scheduler = Scheduler(QUEUE_SIZE, WORKERS_NUM, ADE_TIMEOUT, USE_VIRTUAL_DISPLAY, RESULT_TTL)
    scheduler.start()
    yield scheduler
    scheduler.stop()


def test_engine_normal(launch_engine):
    tasks = [
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'),
        (read_file_binary(os.path.join(TEST_DATA, 'without_password.pka')), None),
    ]

    task_ids = [launch_engine.put(activity, password) for activity, password in tasks]
    sleep_delta = 1
    while len(task_ids):
        time.sleep(sleep_delta)
        for task_id in tuple(task_ids):
            try:
                activity_data = launch_engine.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                task_ids.remove(task_id)
                assert activity_data.data


def test_engine_errors(launch_engine):
    tasks = [
        ('no_such_file.pka'.encode('utf-8'), 'password', ActivityFileReadingError),
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), 'wrong_password', WrongPassword),
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), None, ActivityNeedsPassword),
        (read_file_binary(os.path.join(TEST_DATA, 'topology.pkt')), None, TopologyFilesNotSupported),
    ]

    task_ids = {launch_engine.put(activity, password): error_type for activity, password, error_type in tasks}
    sleep_delta = 1
    while len(task_ids):
        time.sleep(sleep_delta)
        for task_id in tuple(task_ids):
            try:
                activity_data = launch_engine.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                error_type = task_ids.pop(task_id)
                assert isinstance(activity_data.error, error_type)


def test_queue_is_full(launch_engine):
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    tasks = []
    with pytest.raises(QueueIsFull):
        for i in range(QUEUE_SIZE * 2):
            tasks.append(launch_engine.put(activity, password))

    sleep_delta = 1
    while len(tasks):
        time.sleep(sleep_delta)
        for task_id in tuple(tasks):
            try:
                result = launch_engine.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                tasks.remove(task_id)
                assert result.data


def test_task_id_not_found(launch_engine):
    with pytest.raises(TaskIdNotFound):
        launch_engine.get('no_such_task_id')


def test_task_not_completed(launch_engine):
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    task_id = launch_engine.put(activity, password)
    with pytest.raises(TaskNotCompleted):
        launch_engine.get(task_id)


def test_result_ttl():
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    result_ttl = datetime.timedelta(seconds=ADE_TIMEOUT * 2)
    workers_num = 1

    with Scheduler(QUEUE_SIZE, workers_num, ADE_TIMEOUT, USE_VIRTUAL_DISPLAY, result_ttl) as scheduler:
        old_task_id = scheduler.put(activity, password)
        time.sleep(result_ttl.total_seconds())
        young_task_id = scheduler.put(activity, password)
        time.sleep(result_ttl.total_seconds() * 1.5)

        with pytest.raises(TaskIdNotFound):
            scheduler.get(old_task_id)

        assert scheduler.get(young_task_id).data


class Client(threading.Thread):
    def __init__(self, scheduler: Scheduler, tasks: typing.Iterable[typing.Tuple[bytes, str]], result_handler: list):
        super().__init__(daemon=True)
        self._tasks = tasks
        self._scheduler = scheduler
        self._result_handler = result_handler

    def run(self):
        completed_tasks = list()
        running_tasks = [self._scheduler.put(activity, password) for activity, password in self._tasks]
        tasks_num = len(running_tasks)
        while len(completed_tasks) != tasks_num:
            time.sleep(1)
            for task_id in running_tasks:
                try:
                    result = self._scheduler.get(task_id)
                except TaskNotCompleted:
                    pass
                else:
                    completed_tasks.append(result)
                    running_tasks.remove(task_id)

        self._result_handler.extend(completed_tasks)


def test_thread_safe(launch_engine):
    tasks = [
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'),
        (read_file_binary(os.path.join(TEST_DATA, 'without_password.pka')), None),
    ]
    clients_num = WORKERS_NUM * 2
    results = list()
    clients = [Client(launch_engine, tasks, results) for i in range(clients_num)]
    for client in clients:
        client.start()
    for client in clients:
        client.join()

    assert len(results) == len(clients) * len(tasks)


def test_pt_session_restart():
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    queue_size = 20
    tasks_before_session_restart = queue_size // 2
    tasks_num = tasks_before_session_restart + 1
    workers_num = 1

    with Scheduler(
            queue_size,
            workers_num,
            ADE_TIMEOUT,
            USE_VIRTUAL_DISPLAY,
            RESULT_TTL,
            tasks_before_session_restart
    ) as scheduler:
        running_tasks = [scheduler.put(activity, password) for i in range(tasks_num)]
        while len(running_tasks):
            time.sleep(1)
            for task_id in running_tasks:
                try:
                    result = scheduler.get(task_id)
                except TaskNotCompleted:
                    pass
                else:
                    running_tasks.remove(task_id)
                    assert not result.error



import subprocess
import json
import time
import platform
import os
import socket
import typing
import logging
import datetime
import re
import signal

logger = logging.getLogger('adengine.pt')

__all__ = [
    'PacketTracerSession',
    'extract_data',

    'PacketTracerSessionError',
    'PacketTracerSessionNotStarted',
    'PacketTracerSessionAlreadyRunning',

    'LaunchingPacketTracerError',
    'LaunchingPacketTracerTimeout',

    'ExtractorError',
    'GeneralError',
    'WrongCredentials',
    'ConnectionFailed',
    'ArgumentsParsingError',
    'ActivityFileReadingError',
    'ActivityNeedsPassword',
    'TopologyFilesNotSupported',
    'WrongPassword',
]

try:
    PT_HOME = os.environ['PT7HOME']
except KeyError:
    raise Exception('PacketTracer is not installed')

PT_BIN_ROOT = os.path.join(PT_HOME, 'bin')
PT_EXTENSIONS_ROOT = os.path.join(PT_HOME, 'extensions')
PT_PROCESS_NAME = 'PacketTracer7'

if platform.system() == 'Windows':
    PT_EXECUTABLE = os.path.join(PT_BIN_ROOT, 'PacketTracer7.exe')
elif platform.system() == 'Linux':
    PT_EXECUTABLE = os.path.join(PT_BIN_ROOT, 'PacketTracer7')
else:
    raise Exception(f'Unsupported OS: {platform.system()}')

ADE_ROOT = os.path.join(PT_EXTENSIONS_ROOT, 'ade')
ADE = os.path.join(ADE_ROOT, 'ade.jar')
if not os.path.isfile(ADE):
    raise Exception(f'Activity Data Extractor not found in the following path: {ADE}')


def _check_java_version():
    """Checks java version.

    If Java version 8 or newer, do nothing, else, raises Exception.

    Raises:
        Exception: when java version older than 8
    """
    java_ver_result = subprocess.run(['java', '-version'],
                                     text=True,
                                     capture_output=True)
    if java_ver_result.returncode:
        raise Exception('Java not installed. Please install Java 8 or newer.')
    else:
        java_ver = re.search('[0-9]+\.[0-9]+\.[0-9_+]+', java_ver_result.stderr.split('\n')[0]).group(0)
        java_ver = java_ver.split('.')
        if not (int(java_ver[0]) >= 9 or (java_ver[0] == '1' and java_ver[1] == '8')):
            raise Exception(f'Installed Java is too old: {".".join(java_ver)}. Please install Java 8 or newer.')


_check_java_version()


class LaunchingPacketTracerError(Exception):
    """This exception is used to indicate that Packet Tracer stated with error"""

    def __init__(self, port: int, nogui: bool):
        super().__init__()
        self._port = port
        self._nogui = nogui

    def __str__(self):
        return f'An error occurred while launching Packet Tracer on port {self._port} with nogui={self._nogui}'

    @property
    def port(self) -> int:
        return self._port

    @property
    def nogui(self) -> bool:
        return self._nogui


class LaunchingPacketTracerTimeout(LaunchingPacketTracerError):
    """This exception is used to indicate that Packet Tracer reached timeout while starting."""

    def __str__(self):
        return f'Timeout occurred while launching Packet Tracer on port {self._port} with nogui={self._nogui}'


class PortInUse(LaunchingPacketTracerError):
    """This exception is used to indicate that Packet Tracer can not start on the specified port,
    because this port is already in use by another program."""

    def __str__(self):
        return f'Port {self._port} is already in use'


class ExtractorError(Exception):
    """Base extractor error."""
    pass


class GeneralError(ExtractorError):
    """This exception is thrown when ADE failed with unknown error."""
    pass


class WrongCredentials(ExtractorError):
    """This exception is thrown when ADE contains wrong credentials."""
    pass


class ConnectionFailed(ExtractorError):
    """This exception is thrown when ADE failed to connect to Packet Tracer."""
    pass


class ArgumentsParsingError(ExtractorError):
    """This exception is thrown when ADE failed to parse args."""
    pass


class ActivityFileReadingError(ExtractorError):
    """This exception is thrown when ADE failed to read activity file."""
    pass


class ActivityNeedsPassword(ExtractorError):
    """This exception is thrown when activity file needs password, but no password was provided."""
    pass


class TopologyFilesNotSupported(ExtractorError):
    """This exception is thrown when topology file was provided to ADE instead of activity file."""
    pass


class WrongPassword(ExtractorError):
    """This exception is thrown when password that provided for activity file is wrong."""
    pass


class PacketTracerSessionError(Exception):
    """Base Packet Tracer session exception"""

    def __init__(self, port: int, nogui: bool):
        self._port = port
        self._nogui = nogui

    @property
    def port(self) -> int:
        return self._port

    @property
    def nogui(self) -> bool:
        return self._nogui


class PacketTracerSessionAlreadyRunning(PacketTracerSessionError):
    """This exception is thrown when the current session of Packet Tracer is already running and
    user tries to start a new one from the same instance of PacketTracerSession."""

    def __str__(self):
        return f'This instance of PacketTracerSession with port={self._port}, and nogui={self._nogui} is already running'


class PacketTracerSessionNotStarted(PacketTracerSessionError):
    """This exception is thrown when current PacketTracerSession is not stated yet, but user tries to stop it."""

    def __str__(self):
        return f'This instance of PacketTracerSession with port={self._port}, and nogui={self._nogui} is not started'


def _is_port_in_use(port: int) -> bool:
    """Checks if this TCP port in use.

    Args:
        port: number of port to check (from 1 to 65535)

    Returns:
        True - if port is in use, else False
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex(('localhost', port)) == 0


def _get_free_port() -> int:
    """Returns free TCP port.

    Gets random free tcp port and returns it's number.

    Returns:
        Number of free port.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(('localhost', 0))
        return sock.getsockname()[1]


def _make_params(*args, **kwargs) -> typing.Tuple[str, ...]:
    """Makes tuple with arguments to send to subprocess.run() from *args and **kwargs.

    Firstly every arg from *args stored in result tuple same order as provided.\n
    After that every kwarg from **kwargs stored in result tuple with value like: ("--param_name", "param_value")\n
    Only -- is used with arguments provided in **kwargs.\n
    If kwarg's value is None, this kwarg will be skipped.

    Args:
        *args: Arguments to store
        **kwargs: Arguments with values to store

    Examples:
        >>> _make_params(('ls', '-l'), color='always', other_argument=None)
        ('ls', '-l', '--color', 'always')

    Returns:
        A tuple with all arguments from **args and **kwargs
    """
    params = [str(arg) for arg in args]
    for name, value in kwargs.items():
        if value is not None:
            if value:
                params.append(f'--{name}')
                if type(value) != bool:
                    params.append(str(value))
    return tuple(params)


def _check_xvfb_run_installed():
    """Checks if xvfb-run is installed.

    Returns:
        True, if xvfb-run is installed, otherwise False.
    """
    completed_process = subprocess.run(('xvfb-run', '--help'),
                                       stderr=subprocess.DEVNULL,
                                       stdout=subprocess.DEVNULL)
    return completed_process.returncode == 0


def _get_process_pid_by_name(process_name: str) -> typing.Tuple[int, ...]:
    """Gets list of pids of a processes with specified process name.

    If there

    Args:
        process_name: Name of the process.

    Returns:
        Tuple with process pids.
    """
    completed_process = subprocess.run(('pidof', process_name),
                                       capture_output=True,
                                       text=True)
    pids = completed_process.stdout.strip('\n').split()
    return tuple(int(pid) for pid in pids)


def _get_process_command_by_pid(pid: int) -> str or None:
    """Gets process command by pid.

    Args:
        pid: Process pid.

    Returns:
        Process command or None if there is no process with this pid.
    """
    completed_process = subprocess.run(('ps', '-p', str(pid), '-o', 'command'),
                                       capture_output=True,
                                       text=True)
    command = completed_process.stdout.strip('\n').split('\n')[1]  # second row contains command
    return command if command else None


def _get_packet_tracer_pid(port: int) -> int or None:
    """Gets Packet Tracer pid.

    Args:
        port: IPC port.

    Returns:
        Process pid.
    """
    pids = _get_process_pid_by_name(PT_PROCESS_NAME)
    for pid in pids:
        command = _get_process_command_by_pid(pid)
        if command and f'--ipc-port {port}' in command:
            return pid
    return None


def _terminate_process(pid: int, blocking=False):
    """Terminates process with specified pid.
    
    Args:
        pid: Process pid.
        blocking: If set to True, waits for process completion.
    """
    os.kill(pid, signal.SIGTERM)
    if blocking:
        os.waitpid(pid, 0)


def _launch_pt(port=39000, nogui=False, timeout: int = -1, use_virtual_display=False) -> int:
    """Launches Packet Tracer.

    Launches Packet Tracer on the specified port.\n
    If nogui is set to True, launches Packet Tracer in nogui mode (some gui objects still materialise).

    Timeout behaviour:
        * If timeout set to -1 (default), waits forever while Packet Tracer would be fully launched.
        * If timeout set to 0, immediately check if Packet Tracer is fully launched and throw an exception if not.
        * If timeout set > 0, waits for timeout seconds, check if Packet Tracer if fully launched and throw
        exception if not.

    Args:
        port: Number of the port to launch Packet Tracer on.
        nogui: If set to True, launches Packet Tracer in nogui mode.
        timeout: Timeout to launch Packet Tracer.
        use_virtual_display: If set to True, start Packet Tracer with gui in virtual display.

    Returns:
        Process pid for launched Packet Tracer.
    Raises:
        PortInUse: when provided TCP port is already in use.
        LaunchingPacketTracerError: when Packet Tracer didn't launch due to unknown error.
        LaunchingPacketTracerTimeout: when there is a timeout.
    """
    if platform.system() == 'Linux':
        ld_library_path = 'LD_LIBRARY_PATH'
        if not os.getenv(ld_library_path):
            os.environ[ld_library_path] = PT_BIN_ROOT

    # check if port is in use
    if _is_port_in_use(port):
        raise PortInUse(port, nogui)

    params = _make_params(PT_EXECUTABLE,
                          **{
                              'ipc-port': port,
                              'nogui': nogui
                          })
    if use_virtual_display:
        if _check_xvfb_run_installed():
            logger.debug('xvfb was found')
            params = _make_params('xvfb-run', '-a') + params
        else:
            logger.warning('xvfb not found, use default display')

    process = subprocess.Popen(params,
                               cwd=PT_BIN_ROOT,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)

    if process.poll():
        raise LaunchingPacketTracerError(port, nogui)

    # wait until PT is up
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def pt_is_up():
        nonlocal sock
        return not sock.connect_ex(('localhost', port))

    if timeout == -1:
        delay = 1
        while True:
            if pt_is_up():
                break
            else:
                time.sleep(delay)

    elif timeout == 0:
        if not pt_is_up():
            _terminate_process(_get_packet_tracer_pid(port))
            raise LaunchingPacketTracerTimeout(port, nogui)

    else:
        time.sleep(timeout)
        if not pt_is_up():
            _terminate_process(_get_packet_tracer_pid(port))
            raise LaunchingPacketTracerTimeout(port, nogui)

    logger.debug(f'Packet Tracer is up on port {port}')
    return _get_packet_tracer_pid(port)


def _extract_data(port: int, filepath: str, password: str = None, timeout=10, conn_attempts=5, conn_delay=100):
    """Extracts data from the activity file.

    This is a wrapper for ADE. It calls ADE and parses resulted JSON.

    Args:
        port: Number of TCP port to connect to Packet Tracer.
        filepath: Path to activity file.
        password: Password for the provided activity file.
        timeout: Timeout for ADE - time to wait when the message "Can not read this file" appears before
        decide that it is not an activity file (in seconds).
        conn_attempts: Number of connection to perform.
        conn_delay: Delay between connection attempts (in milliseconds).

    Returns:
        Dict which is parsed JSON from ADE.
    Raises:
        GeneralError: when an unknown error occurred in ADE.
        WrongCredentials: when ADE has wrong credentials and can not connect to Packet Tracer.
        ConnectionFailed: when ADE failed to establish connection with Packet Tracer.
        ArgumentsParsingError: when ADE failed to parse arguments.
        ActivityFileReadingError: when ADE can not read activity file.
        WrongPassword: when wrong password was provided to specified activity.
        ActivityNeedsPassword: when no password was provided, but specified activity requires it.
        TopologyFilesNotSupported: when topology file was provided instead of activity file.
    """
    params = _make_params('java', '-Dfile.encoding=utf-8', '-jar', ADE,
                          input=filepath,
                          key=password,
                          server='localhost',
                          port=port,
                          conn_attempts=conn_attempts,
                          conn_delay=conn_delay,
                          timeout=timeout * 1000)

    before = datetime.datetime.now()
    completed_process = subprocess.run(params,
                                       capture_output=True,
                                       text=True)
    after = datetime.datetime.now()
    logger.debug(f'Extracted data from {filepath}. Consumed time: {after - before}')

    error_mapping = {
        1: GeneralError,
        2: WrongCredentials,
        3: ConnectionFailed,
        4: ArgumentsParsingError,
        5: ActivityFileReadingError,
        6: None,  # an error occurred while writing data to file
        7: WrongPassword,
        8: ActivityNeedsPassword,
        9: TopologyFilesNotSupported,
    }

    if completed_process.returncode:
        error = error_mapping[completed_process.returncode]
        raise error(completed_process.stderr)

    for line in completed_process.stdout.split(os.linesep):
        first_bracket_index = line.find('{')
        last_bracket_index = line.rfind('}')
        if first_bracket_index != -1 and last_bracket_index != -1:
            try:
                return json.loads(line[first_bracket_index:last_bracket_index + 1])
            except json.JSONDecodeError:
                pass


class PacketTracerSession:
    """This class represents the Packet Tracer session.

    If no port is provided, random free TCP port will be used.

    Call start() to launch the Packet Tracer session and stop() to stop it.\n

    This class can be used as context manager:
        * When enter context - start() method is used.
        * When leave context - stop() method is used.

    Typical usage: as a parameter for extract_data() function to extract data from activity file.

    Examples:
        >>> with PacketTracerSession() as session:
        >>>     extract_data(session, 'activity.pka', 'password')
    """

    def __init__(self, port: int = 0, nogui=False, use_virtual_display=False):
        self._port = port
        self._nogui = nogui
        self._use_virtual_display = use_virtual_display
        self._pid: int = None

    def __enter__(self) -> 'PacketTracerSession':
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __repr__(self):
        return f'{self.__class__}({self._port, self._nogui})'

    def start(self):
        # if port not provided choose random port
        self._port = self._port or _get_free_port()

        if not self._pid:
            logger.debug(f'Starting Packet Tracer session on port {self._port}')
            self._pid = _launch_pt(self._port, self._nogui, use_virtual_display=self._use_virtual_display)
        else:
            raise PacketTracerSessionAlreadyRunning(self._port, self._nogui)

    def stop(self):
        if self._pid:
            logger.debug(f'Stop Packet Tracer on port {self._port}')
            _terminate_process(self._pid)
            self._pid = None
        else:
            raise PacketTracerSessionNotStarted(self._port, self._nogui)

    @property
    def port(self) -> int:
        return self._port

    @property
    def running(self):
        return True if self._pid else False


def extract_data(session: PacketTracerSession, filepath: str, password: str = None, timeout=10) -> dict:
    """Extracts data from the activity file.

    This is a wrapper for ADE. It calls ADE and parses resulted JSON.

    Args:
        session: PacketTracerSession to connect to.
        filepath: Path to activity file.
        password: Password for the provided activity file.
        timeout: Timeout for ADE - time to wait when the message "Can not read this file" appears before
        decide that it is not an activity file (in seconds).

    Returns:
        Dict which is parsed JSON from ADE.
    Raises:
        GeneralError: when an unknown error occurred in ADE.
        WrongCredentials: when ADE has wrong credentials and can not connect to Packet Tracer.
        ConnectionFailed: when ADE failed to establish connection with Packet Tracer.
        ArgumentsParsingError: when ADE failed to parse arguments.
        ActivityFileReadingError: when ADE can not read activity file.
        WrongPassword: when wrong password was provided to specified activity.
        ActivityNeedsPassword: when no password was provided, but specified activity requires it.
        TopologyFilesNotSupported: when topology file was provided instead of activity file.
    """
    return _extract_data(session.port, filepath, password, timeout)


if __name__ == '__main__':
    print('This is module with functions to extract data from activity file.')



import dataclasses
import multiprocessing as mp
import os
import datetime
import uuid
import tempfile
import signal
import logging
import threading
import time

from .pt import *


logger = logging.getLogger('adengine.engine')


__all__ = [
    'Task',
    'ActivityData',

    'Scheduler',

    'SchedulerError',
    'TaskIdNotFound',
    'TaskNotCompleted',
    'QueueIsFull',
]


@dataclasses.dataclass(frozen=True)
class Task:
    """This class represents the task for Scheduler."""
    id: str
    activity: bytes
    password: str


@dataclasses.dataclass(frozen=True)
class ActivityData:
    """This class represents the result of extracting data."""
    consumed_time: datetime.timedelta
    error: ExtractorError
    data: dict


class Worker:
    """Worker for extract data in an isolated process.

    All logic is stored in __call__ method. The Worker starts Packet tracer session and blocks until reading the new task
    from the queue after that worker extracts data from the activity file and push the result to the resulting pool.
    If any error occurred while extraction, it is written to the corresponding attribute in ActivityData instance
    and the data attribute is set to None.

    When __call__ executes, it creates a temporary file to write activity binaries to in specified working directory.
    By default, it is strongly recommended to set the temporary directory in your OS as working directory every time,
    except debugging purposes. Temporary directory depends on your OS:
    /tmp in Linux, C:\\Users\\<username>\\AppData\\Local\\Temp in Windows.

    When a process with running Worker terminated, Worker erases temporary file and terminate Packet Tracer process.
    """
    def __init__(self, queue: mp.Queue, result_pool: dict, work_dir: str, extract_timeout: int,
                 use_virtual_display=False, tasks_before_session_restart=100):
        """Initialize Worker

        Args:
            queue: Queue to get tasks from.
            result_pool: Result pool to store results after extracting.
            work_dir: Working directory. Strongly recommend to set the temporary directory of your OS, or it's subdirectory.
            extract_timeout: Timeout for ADE to extract data. This parameter means the same as same parameter in pt.extract_data().
            use_virtual_display: If set to True, start Packet Tracer with gui in virtual display.
            tasks_before_session_restart: Number of tasks before restarting PacketTracer session.
        """
        self._queue = queue
        self._result_pool = result_pool
        self._work_dir = work_dir
        self._extract_timeout = extract_timeout
        self._use_virtual_display = use_virtual_display
        self._tasks_before_session_restart = tasks_before_session_restart

    def __repr__(self):
        args = (self._queue, self._result_pool, self._work_dir, self._extract_timeout, self._use_virtual_display)
        return f'{self.__class__.__name__}({", ".join(obj.__repr__() for obj in args)})'

    def __call__(self, *args, **kwargs):
        """Method with main Worker logic.

        Executes extracting data from activity files.\n
        All arguments provided for this method are ignored.

        Args:
            *args: unused args
            **kwargs: unused kwargs
        """
        logger.debug(f'Starting worker {os.getpid()}')

        signal.signal(signal.SIGTERM, self._stop)

        self._filepath = os.path.join(self._work_dir, f'activity{os.getpid()}.pka')

        task_counter = 0

        self._start_session()

        while True:
            if task_counter == self._tasks_before_session_restart:
                self._stop_session()
                self._start_session()
                task_counter = 0

            task = self._queue.get(True)
            task_counter += 1
            logger.debug(f'Worker {os.getpid()} reserved task {task.id}')

            with open(self._filepath, 'wb') as file:
                file.write(task.activity)

            before = datetime.datetime.now()
            try:
                data = extract_data(self._session, self._filepath, task.password, self._extract_timeout)
            except ExtractorError as e:
                error = e
                data = None
            else:
                error = None
            after = datetime.datetime.now()
            time_consumed = after - before

            self._result_pool[task.id] = ActivityData(time_consumed, error, data), after
            logger.debug(f'Task {task.id} completed. Time consumed: {time_consumed}')

    def _clean(self):
        """Cleans all temporary files and directories, created during executing."""
        if os.path.exists(self._filepath):
            os.remove(self._filepath)
            logger.debug(f'Worker {os.getpid()} removed temporary file')

    def _stop(self, signum, frame):
        """Stops Worker on capture SIGTERM signal."""
        del signum, frame  # unused
        logger.debug(f'Stop worker {os.getpid()}...')
        try:
            self._stop_session()
            self._clean()
        except Exception as e:
            logger.debug(f'Worker {os.getpid()} stopped with error: {e}')
            exit(1)
        else:
            logger.debug(f'Worker {os.getpid()} stopped correctly')
            exit(0)

    def _stop_session(self):
        """Stops PacketTracer session"""
        if self._session and self._session.running:
            self._session.stop()

    def _start_session(self):
        """Starts new PacketTracer session"""
        self._session = PacketTracerSession(use_virtual_display=self._use_virtual_display)
        self._session.start()


class SchedulerError(Exception):
    """Base Scheduler exception."""
    pass


class TaskIdNotFound(SchedulerError):
    """This exception is thrown when task id not found in result pool or in not completed tasks."""
    pass


class TaskNotCompleted(SchedulerError):
    """This exception is thrown when task with requested task id not completed."""
    pass


class QueueIsFull(SchedulerError):
    """This exception is thrown when user it trying to put new task to Scheduler with full queue."""
    pass


class Scheduler:
    """Class to control Workers.

    This class is used to control Workers. Scheduler lifecycle:
        1. Creates a queue for tasks with max size specified in queue_size.
        2. Creates result pool to store results of extraction.
        3. Creates a pool of workers with the size specified in workers_num.
        4. Creates subdirectory in OS temporary directory.

    On-call start() method, Scheduler start all Workers in the pool in different processes.\n
    On-call stop() method, Scheduler terminates all Workers processes.

    This class can be used as context manager:
        * When enter context - start() method is used.
        * When leave context - stop() method is used.
    """
    def __init__(self, queue_size: int, workers_num: int, extract_timeout: int, use_virtual_display=False,
                 result_ttl=datetime.timedelta(minutes=5), tasks_before_session_restart=100):
        """Initialize Scheduler.

        Args:
            queue_size: Size of the queue with tasks.
            workers_num: Number of workers.
            extract_timeout: Timeout for ADE to extract data. This parameter means the same as same parameter in pt.extract_data().
            use_virtual_display: If set to True, start Packet Tracer with gui in virtual display.
            result_ttl: Time to wait before delete result.
            tasks_before_session_restart: Number of tasks before restarting PacketTracer session.
        """
        self._queue_size = queue_size
        self._queue = mp.Queue(queue_size)
        self._manager = mp.Manager()
        self._result_pool = self._manager.dict()
        self._not_completed_tasks = set()
        self._extract_timeout = extract_timeout
        self._use_virtual_display = use_virtual_display
        self._result_ttl = result_ttl
        self._tasks_before_restart = tasks_before_session_restart

        self._working_dir = os.path.join(tempfile.gettempdir(), 'adengine')
        os.makedirs(self._working_dir, exist_ok=True)

        self._workers = [mp.Process(
            name=f'adengine_worker_{i}',
            target=Worker(self._queue, self._result_pool, self._working_dir, self._extract_timeout,
                          self._use_virtual_display, self._tasks_before_restart)
        ) for i in range(workers_num)]

        self._put_lock = threading.Lock()
        self._get_lock = threading.Lock()

        self._is_alive = False

    def __repr__(self):
        args = (
            self._queue_size,
            len(self._workers),
            self._extract_timeout,
            self._use_virtual_display,
            self._tasks_before_restart
        )
        return f'{self.__class__.__name__}({", ".join(obj.__repr__() for obj in args)})'

    def start(self):
        logger.debug('Starting scheduler...')
        for _worker in self._workers:
            _worker.start()
        threading.Thread(target=self._clean_result_pool,
                         name='ResultPoolCleaner',
                         daemon=True).start()
        self._is_alive = True

    def stop(self):
        logger.debug('Stop scheduler...')
        for worker in self._workers:
            worker.terminate()
        self._manager.shutdown()
        self._is_alive = False
        logger.debug('Scheduler stopped')

    def put(self, activity: bytes, password: str) -> str:
        """Puts task in queue.

        Puts task instance in queue. If the queue is full, raises QueueIsFull exception.

        Args:
            activity: Binary representation of activity file.
            password: Password for activity.

        Returns:
            If the task is created correctly, returns task id.

        Raises:
            QueueIsFull: when try to put task in full queue.
        """
        with self._put_lock:
            if self._queue.full():
                logger.debug(f'Queue of Scheduler {os.getpid()} is full')
                raise QueueIsFull()

            task_id = uuid.uuid1().hex
            self._queue.put(Task(task_id, activity, password))  # blocking safe process operation
            self._not_completed_tasks.add(task_id)
            logger.debug(f'Created task {task_id}')
            return task_id

    def get(self, task_id: str) -> ActivityData:
        """Gets extraction result from result pool.

        Tries to get a task with provided task id from result poll.\n
        If the task is in work, raises TaskInProgress exception.\n
        If the task is not found in result pool and not on progress, it is considered as unknown task and Scheduler raises
        TaskIdNotFound exception.

        Args:
            task_id: Task ID to get.

        Returns:
            An instance of ActivityData for requested task id.

        Raises:
            TaskNotFoundError: when task id not found.
            TaskNotCompleted: when task is not completed yet.
        """
        with self._get_lock:
            if task_id not in self._not_completed_tasks:
                logger.debug(f'Task {task_id} not found')
                raise TaskIdNotFound()
            elif task_id not in self._result_pool:
                logger.debug(f'Task {task_id} not completed')
                raise TaskNotCompleted()
            else:
                self._not_completed_tasks.remove(task_id)
                logger.debug(f'Return task {task_id}')
                return self._result_pool.pop(task_id)[0]

    def _clean_result_pool(self):
        while True:
            logger.debug(f'ResultPoolCleaner sleep for {self._result_ttl}...')
            time.sleep(self._result_ttl.total_seconds())
            logger.debug(f'Start cleaning...')

            if not self._is_alive:
                return

            with self._get_lock:
                results = tuple(self._result_pool.items())

            curr_time = datetime.datetime.now()
            for task_id, (_, finish_time) in results:
                if curr_time - finish_time > self._result_ttl:
                    logger.debug(f'Task {task_id} reached ttl')
                    try:
                        self.get(task_id)
                    except TaskIdNotFound:
                        pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



import os
import argparse
import datetime
import socket
import logging
import traceback

from .engine import Scheduler, TaskNotCompleted, TaskIdNotFound, QueueIsFull
from .messages import *

logger = logging.getLogger('adengine.server')

ADENGINE_UNIX_SOCK = os.environ.get('ADENGINE_UNIX_SOCK', None) or DEFAULT_UNIX_SOCKET


def parse_timedelta(time_str: str):
    """Converts string hh:mm:ss to timedelta."""
    parts = time_str.split(':')
    time_params = dict(zip(['hours', 'minutes', 'seconds'],
                           (int(part) for part in parts)))
    return datetime.timedelta(**time_params)


def parse_args() -> dict:
    parser = argparse.ArgumentParser('ADEngine server')
    parser.add_argument('-qs', '--queue-size',
                        help='Maximum size of the queue with tasks',
                        dest='queue_size',
                        type=int,
                        default=20)
    parser.add_argument('-wn', '--workers-num',
                        help='Number of workers',
                        dest='workers_num',
                        type=int,
                        default=1)
    parser.add_argument('-et', '--extract-timeout',
                        help='Timeout for data extraction in seconds',
                        dest='extract_timeout',
                        type=int,
                        default=5)
    parser.add_argument('-vd', '--virtual-display',
                        help='Use virtual display',
                        dest='virtual_display',
                        action='store_true')
    parser.add_argument('-rt', '--result-ttl',
                        help='Result ttl like hh:mm:ss',
                        dest='result_ttl',
                        type=parse_timedelta,
                        default=datetime.timedelta(minutes=5))
    parser.add_argument('-tb', '--tasks-before-session-restart',
                        help='Number of tasks before restarting PacketTracer session',
                        dest='tasks_before_session_restart',
                        type=int,
                        default=100)
    parser.add_argument('-us', '--unix-socket',
                        help='Path to unix socket',
                        dest='unix_socket',
                        default=ADENGINE_UNIX_SOCK)
    parser.add_argument('-mc', '--max-connections',
                        help='Maximum number of connections to server',
                        dest='max_connections',
                        type=int,
                        default=10)
    return vars(parser.parse_args())


def main():
    args = parse_args()

    unix_socket_path = args['unix_socket']
    if os.path.exists(unix_socket_path):
        logger.info(f'Unix socket {unix_socket_path} already_exists, remove it')
        os.remove(unix_socket_path)

    os.makedirs(os.path.dirname(unix_socket_path), exist_ok=True)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(unix_socket_path)

    returncode = 0

    logger.info('Starting scheduler...')
    with Scheduler(
            args['queue_size'],
            args['workers_num'],
            args['extract_timeout'],
            args['virtual_display'],
            args['result_ttl'],
            args['tasks_before_session_restart'],
    ) as scheduler:

        sock.listen(args['max_connections'])
        logger.info(f'Listening on unix socket: {unix_socket_path}...')
        while True:
            try:
                conn, _ = sock.accept()
                try:
                    message = recv(conn)
                except RecvError as e:
                    logger.error(f'An error occurred while receiving message:\n'
                                 f'{traceback.format_exception(RecvError, e, e.__traceback__)}')

                if isinstance(message, TaskMessage):
                    try:
                        task_id = scheduler.put(message.activity, message.password)
                    except QueueIsFull:
                        send(conn, QueueIsFullMessage())
                    else:
                        send(conn, TaskIdMessage(task_id))

                elif isinstance(message, TaskIdMessage):
                    try:
                        result = scheduler.get(message.task_id)
                    except TaskIdNotFound:
                        send(conn, TaskIdNotFoundMessage())
                    except TaskNotCompleted:
                        send(conn, TaskNotCompletedMessage())
                    else:
                        send(conn, ResultMessage(result))

                conn.close()

            except KeyboardInterrupt:
                break
            except Exception as e:
                logger.error(f'Unexpected exception:\n'
                             f'{traceback.format_exception(e.__class__, e, e.__traceback__)}')
                break

        logger.info('Stop server...')

    return returncode


if __name__ == '__main__':
    exit(main())



import os

TEST_ROOT = os.path.dirname(__file__)
TEST_DATA = os.path.join(TEST_ROOT, 'data')

USE_VIRTUAL_DISPLAY = True


def read_file_binary(filepath: str) -> bytes:
    with open(filepath, 'rb') as file:
        return file.read()



import abc
import base64
import json
import datetime
import logging
import socket
import os
import tempfile

from .engine import ActivityData
from .pt import TopologyFilesNotSupported, WrongPassword, ActivityFileReadingError, ActivityNeedsPassword, \
    ExtractorError

__all__ = [
    'Message',
    'TaskMessage',
    'TaskIdMessage',
    'TaskIdNotFoundMessage',
    'TaskNotCompletedMessage',
    'QueueIsFullMessage',
    'ResultMessage',

    'send',
    'recv',

    'MessageError',
    'RecvError',
    'SendError',

    'DEFAULT_UNIX_SOCKET',
]

logger = logging.getLogger(f'adengine.messages')

BUFFER_SIZE = 1024
DEFAULT_UNIX_SOCKET = os.path.join(tempfile.gettempdir(), 'adengine/adengine.socket')


class MessageError(Exception):
    """Basic message exception."""
    pass


class SendError(Exception):
    """An error occured while sending message."""
    pass


class RecvError(Exception):
    """An error occurred while receiving message."""
    pass


class Message(abc.ABC):
    """Basic class representing message."""

    @abc.abstractmethod
    def _to_json(self) -> dict:
        """Converts object data to json."""
        pass

    def to_json(self) -> dict:
        """Combines _to_json return value with message type."""
        return {
            'message_type': self.__class__.__name__,
            'data': self._to_json(),
        }

    @classmethod
    @abc.abstractmethod
    def _from_json(cls, data: dict) -> 'Message':
        """Converts json to object data."""
        pass

    @classmethod
    def from_json(cls, data: dict) -> 'Message':
        """Retrieves json with object data from all json."""
        return cls._from_json(data['data'])

    def to_bytes(self):
        """Converts object to bytes."""
        return json.dumps(self.to_json()).encode('utf-8')

    @classmethod
    def from_bytes(cls, bytes_string: bytes) -> 'Message':
        """Creates object from bytes."""
        return cls.from_json(json.loads(bytes_string.decode('utf-8')))


class TaskMessage(Message):
    """Class representing message with task."""

    def __init__(self, activity: bytes, password: str or None):
        self._activity = activity
        self._password = password

    def _to_json(self) -> dict:
        return {
            'password': self._password,
            'activity': base64.b64encode(self._activity).decode('utf-8'),
        }

    @classmethod
    def _from_json(cls, data: dict) -> 'TaskMessage':
        return cls(
            base64.b64decode(data['activity'].encode('utf-8')),
            data['password'],
        )

    @property
    def activity(self) -> bytes:
        return self._activity

    @property
    def password(self) -> str:
        return self._password


class TaskIdNotFoundMessage(Message):
    """Class representing message with label that task id not found."""

    def _to_json(self) -> dict:
        return {}

    @classmethod
    def _from_json(cls, data: dict) -> 'TaskIdNotFoundMessage':
        return TaskIdNotFoundMessage()


class TaskNotCompletedMessage(Message):
    """Class representing message with label that task not competed."""

    def _to_json(self) -> dict:
        return {}

    @classmethod
    def _from_json(cls, data: dict) -> 'TaskNotCompletedMessage':
        return TaskNotCompletedMessage()


class QueueIsFullMessage(Message):
    """Class representing message with label that queue is full."""

    def _to_json(self) -> dict:
        return {}

    @classmethod
    def _from_json(cls, data: dict) -> 'QueueIsFullMessage':
        return QueueIsFullMessage()


class UnknownMessage(Message):
    """
    Class representing message which is send to client if received message
    is valid, but has not supported message type.
    """

    def _to_json(self) -> dict:
        return {}

    @classmethod
    def _from_json(cls, data: dict) -> 'UnknownMessage':
        return UnknownMessage()


class TaskIdMessage(Message):
    """Class representing message with task id."""

    def __init__(self, task_id: str):
        self._task_id = task_id

    def _to_json(self) -> dict:
        return {
            'task_id': self._task_id,
        }

    @classmethod
    def _from_json(cls, data: dict) -> 'TaskIdMessage':
        return cls(data['task_id'])

    @property
    def task_id(self) -> str:
        return self._task_id


class ResultMessage(Message):
    """Class representing result of task completion."""

    def __init__(self, activity_data: ActivityData):
        self._activity_data = activity_data

    def _to_json(self) -> dict:
        return {
            'consumed_time': {
                'seconds': self._activity_data.consumed_time.seconds,
                'microseconds': self._activity_data.consumed_time.microseconds,
            },
            'error': self._activity_data.error.__class__.__name__ if self._activity_data.error else None,
            'data': self._activity_data.data,
        }

    @classmethod
    def _from_json(cls, data: dict) -> 'ResultMessage':
        return cls(
            ActivityData(
                datetime.timedelta(
                    seconds=data['consumed_time']['seconds'],
                    microseconds=data['consumed_time']['microseconds'],
                ),
                cls._str2error(data['error']),
                data['data'],
            )
        )

    @property
    def activity_data(self) -> ActivityData:
        return self._activity_data

    @staticmethod
    def _error2str(error: ExtractorError) -> str:
        """Converts extractor error to string"""
        return error.__class__.__name__

    @staticmethod
    def _str2error(string: str) -> ExtractorError:
        """Converts string to extractor error"""
        errors = {error.__name__: error for error in (
            TopologyFilesNotSupported,
            WrongPassword,
            ActivityFileReadingError,
            ActivityNeedsPassword,
        )}
        return errors[string]() if string else None


def bytes2message(bytes_string: bytes):
    """Creates Message subclass from bytes."""
    messages = {message.__name__: message for message in (
        TaskMessage,
        TaskIdMessage,
        TaskIdNotFoundMessage,
        TaskNotCompletedMessage,
        QueueIsFullMessage,
        ResultMessage,
    )}
    message_type = json.loads(bytes_string.decode('utf-8'))['message_type']
    return messages[message_type].from_bytes(bytes_string)


def send(sock: socket.socket, message: Message):
    """Sends message using socket."""
    try:
        data = message.to_bytes()
        logger.debug(f'SEND data length: {len(data)}')
        data = len(data).to_bytes(8, byteorder='big', signed=False) + data
        sock.sendall(data)
        logger.debug(f'Sent message {message.__class__.__name__}')
    except Exception as e:
        raise SendError(e)


def recv(sock: socket.socket) -> bytes:
    """Receives message using socket."""
    try:
        total_bytes_num = int.from_bytes(sock.recv(8), byteorder='big', signed=False)
        logger.debug(f'RECEIVE data length: {total_bytes_num}')
        data = bytes()
        while len(data) != total_bytes_num:
            data += (sock.recv(BUFFER_SIZE))
        message = bytes2message(data)
        logger.debug(f'Received message {message.__class__.__name__}')
        return message
    except Exception as e:
        raise RecvError(e)



import pytest
import sys
import subprocess
import signal
import datetime
import os
import time
import tempfile
import multiprocessing as mp
import itertools

from adengine.client import Client
from adengine.engine import TaskIdNotFound, TaskNotCompleted, QueueIsFull, ActivityData
from adengine.pt import WrongPassword, ActivityFileReadingError, ActivityNeedsPassword, TopologyFilesNotSupported

from test import TEST_DATA, USE_VIRTUAL_DISPLAY, read_file_binary

QUEUE_SIZE = 10
WORKERS_NUM = 2
ADE_TIMEOUT = 5
RESULT_TTL = datetime.timedelta(hours=1)
TASKS_BEFORE_SESSION_RESTART = 100
MAX_CONNECTIONS = QUEUE_SIZE
UNIX_SOCKET = os.path.join(tempfile.gettempdir(), 'adengine_test/adengine.socket')


@pytest.fixture(scope='module', autouse=True)
def server():
    cmd = [
        sys.executable,
        '-m', 'adengine.server',
        '--queue-size', str(QUEUE_SIZE),
        '--workers-num', str(WORKERS_NUM),
        '--extract-timeout', str(ADE_TIMEOUT),
        '--virtual-display' if USE_VIRTUAL_DISPLAY else '',
        '--result-ttl', str(RESULT_TTL),
        '--tasks-before-session-restart', str(TASKS_BEFORE_SESSION_RESTART),
        '--max-connections', str(MAX_CONNECTIONS),
        '--unix-socket', UNIX_SOCKET,
    ]
    process = subprocess.Popen(cmd)
    time.sleep(1)
    yield
    process.send_signal(signal.SIGINT)
    process.wait()


@pytest.fixture(scope='function')
def client():
    return Client(UNIX_SOCKET)


def test_client_normal(client):
    tasks = [
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'),
        (read_file_binary(os.path.join(TEST_DATA, 'without_password.pka')), None),
    ]

    task_ids = [client.put(activity, password) for activity, password in tasks]
    sleep_delta = 1
    while len(task_ids):
        time.sleep(sleep_delta)
        for task_id in tuple(task_ids):
            try:
                activity_data = client.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                task_ids.remove(task_id)
                assert activity_data.data


def test_client_errors(client):
    tasks = [
        ('no_such_file.pka'.encode('utf-8'), 'password', ActivityFileReadingError),
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), 'wrong_password', WrongPassword),
        (read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), None, ActivityNeedsPassword),
        (read_file_binary(os.path.join(TEST_DATA, 'topology.pkt')), None, TopologyFilesNotSupported),
    ]

    task_ids = {client.put(activity, password): error_type for activity, password, error_type in tasks}
    sleep_delta = 1
    while len(task_ids):
        time.sleep(sleep_delta)
        for task_id in tuple(task_ids):
            try:
                activity_data = client.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                error_type = task_ids.pop(task_id)
                assert isinstance(activity_data.error, error_type)


def test_queue_is_full(client):
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    tasks = []
    with pytest.raises(QueueIsFull):
        for i in range(QUEUE_SIZE * 2):
            tasks.append(client.put(activity, password))

    sleep_delta = 1
    while len(tasks):
        time.sleep(sleep_delta)
        for task_id in tuple(tasks):
            try:
                result = client.get(task_id)
            except TaskNotCompleted:
                pass
            else:
                tasks.remove(task_id)
                assert result.data


def test_task_id_not_found(client):
    with pytest.raises(TaskIdNotFound):
        client.get('no_such_task_id')


def test_task_not_completed(client):
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    task_id = client.put(activity, password)
    with pytest.raises(TaskNotCompleted):
        client.get(task_id)


def extract_data(activity: bytes, password: str or None) -> ActivityData:
    client = Client(UNIX_SOCKET)
    task_id = client.put(activity, password)
    while True:
        time.sleep(1)
        try:
            result = client.get(task_id)
        except TaskNotCompleted:
            continue
        else:
            return result


def test_process_safe():
    activity, password = read_file_binary(os.path.join(TEST_DATA, 'with_password.pka')), '123'
    tasks_num = WORKERS_NUM * 4

    with mp.Pool(WORKERS_NUM) as pool:
        result = pool.starmap(extract_data, itertools.repeat((activity, password), tasks_num))

    for activity_data in result:
        assert activity_data.data



import pytest
import os
import unittest.mock as mock

from adengine.pt import *

from test import TEST_DATA, USE_VIRTUAL_DISPLAY


@pytest.fixture(scope='module')
def launch_packet_tracer_session():
    session = PacketTracerSession(use_virtual_display=USE_VIRTUAL_DISPLAY)
    session.start()
    yield session
    session.stop()


@pytest.mark.parametrize(
    ['filepath', 'password'],
    [
        pytest.param(
            os.path.join(TEST_DATA, 'with_password.pka'), '123', id='with_password'
        ),
        pytest.param(
            os.path.join(TEST_DATA, 'without_password.pka'), None, id='without_password'
        ),
    ]
)
def test_extract_data(filepath, password, launch_packet_tracer_session):
    extract_data(launch_packet_tracer_session, filepath, password)


@pytest.mark.parametrize(
    ['filepath', 'password', 'exception'],
    [
        pytest.param(
            os.path.join(TEST_DATA, 'no_such_file.pka'), None, ActivityFileReadingError, id='ActivityFileReadingError'
        ),
        pytest.param(
            os.path.join(TEST_DATA, 'with_password.pka'), 1234, WrongPassword, id='WrongPassword'
        ),
        pytest.param(
            os.path.join(TEST_DATA, 'with_password.pka'), None, ActivityNeedsPassword, id='ActivityNeedsPassword'
        ),
        pytest.param(
            os.path.join(TEST_DATA, 'topology.pkt'), None, TopologyFilesNotSupported, id='TopologyFilesNotSupported'
        ),
    ]
)
def test_extract_data_exceptions(filepath, password, exception, launch_packet_tracer_session):
    with pytest.raises(exception):
        extract_data(launch_packet_tracer_session, filepath, password)


def test_connection_failed(launch_packet_tracer_session):
    with mock.patch('adengine.pt.PacketTracerSession.port', new_callable=lambda: 80):
        with pytest.raises(ConnectionFailed):
            extract_data(launch_packet_tracer_session, os.path.join(TEST_DATA, 'without_password.pka'), None)



import logging
import sys

logger = logging.getLogger('adengine')
logger.setLevel(logging.INFO)

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
handler.setLevel(logging.DEBUG)

logger.addHandler(handler)


__all__ = [
    'pt',
    'engine',
]
